/**
 * Task that decodes.
 */

#include <string.h>
#include <limits.h>
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/family/c66/Cache.h>

#include "H264HighProfileDecoder.h"
#include "BufferDescriptorManager.h"
#include "Heaps.h"
#include "Configuration.h"
#include "GlobalGates.h"
#include "DecodeTaskCommon.h"

#define EDMA_SIZE_ALIGNMENT (0x400)
#define MAX_BUFFER_IDS_TO_FREE (16)

extern cregister volatile unsigned int DNUM;
extern cregister volatile unsigned int TSCH;
extern cregister volatile unsigned int TSCL;

static const int8_t* const inputBufferStartAddress = (int8_t*) 0x90000000;
static const int8_t* const inputBufferEndAddress = (int8_t*) (0x90000000 + 1024 * 1024 * 5 / 2);

static const int8_t* const outputBufferStartAddress = (int8_t*) 0x98000000;

static ResultCode doMasterDecoding(int decoderIndex, DecoderComponent* decoderComponent, int maxEncodersCount);

void decodeMasterTask(UArg decoderIndex, UArg arg1)
{
    const DspConfiguration* configuration = Configuration_get();
    const DspDecoderConfiguration* decoderConfiguration = &configuration->decoderConfigurations[decoderIndex];
    DecoderComponent* decoderComponent = NULL;
    int encodersCount = 1;

    System_printf("Create decoder component\n");
    System_flush();

    decoderComponent = (DecoderComponent*) TranscodeComponent_create(sizeof(DecoderComponent), TranscodeComponentType_Decoder, decoderConfiguration->coreIds, Common_getDecoderAlgorithmIndex(decoderIndex));

    if (decoderComponent == NULL) {
        System_printf("Failed to create decoder component\n");
        goto TASK_END;
    }

    while (1) {
        if (TranscodeComponent_waitErrorFromInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES) != ResultCode_Ok) {
            System_printf("Error in decoder on another core\n");
            goto TASK_END;
        }

        ResultCode error = doMasterDecoding(decoderIndex, decoderComponent, encodersCount);

        switch (error) {
        case ResultCode_Error:
            System_printf("Fatal master decoder task error. Terminating task.\n");
            goto TASK_END;

        case ResultCode_ErrorNonFatal:
            System_printf("Non fatal master decoder task error. Restarting task.\n");
            break;

        default:
            System_printf("Unknown master decoder task error. Terminating task.\n");
            goto TASK_END;
        }

        System_flush();

        TranscodeComponent_waitInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES);
        TranscodeComponent_reset(&decoderComponent->transcodeComponent);
        TranscodeComponent_waitInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES);
    }

TASK_END:

    System_flush();

    if (decoderComponent != NULL) {
        TranscodeComponent_delete(decoderComponent->transcodeComponent.id);
    }
}

static ResultCode doMasterDecoding(int decoderIndex, DecoderComponent* decoderComponent, int maxEncodersCount)
{
    int8_t* inputBufferPointer = (int8_t*) inputBufferStartAddress;
    int8_t* outputBufferPointer = (int8_t*) outputBufferStartAddress;

    int i;
    int result;
    ResultCode resultError = ResultCode_Error;
    ResultCode multicoreErrorToSet = ResultCode_Error;
    Error_Block eb;
    Decoder* decoder = NULL;
    IVIDDEC2_Handle decoderHandle;
    IVIDDEC2_Fxns* decoderFunctions;
    IVIDDEC2_DynamicParams *decoderDynamicParams;
    IVIDDEC2_Status* decoderStatus;
    IVIDDEC2_InArgs* inDecoderArguments;
    IVIDDEC2_OutArgs* outDecoderArguments;

    XDM1_BufDesc inDecoderBufferDescriptor = { 0 };
    XDM_BufDesc* outDecoderBufferDescriptor = NULL;
    int inputDecoderBufferSize;
    int8_t* inputDecoderBuffer = NULL;
    BufferDescriptorManagerHandle decoderBufferManager = NULL;
    int bytesAvailable = 0;
    int isNewInputFrameNeeded = 1;
    const DspConfiguration* configuration = Configuration_get();
    const DspDecoderConfiguration* decoderConfiguration = &configuration->decoderConfigurations[decoderIndex];
    IArg gateKey;

    // Shared data, synchronization, multi-core
    int isMulticore = (decoderComponent->transcodeComponent.numberOfInstances > 1);
    GateMP_Handle transcodeComponentGate = GlobalGates_getTranscodeComponentGate();

    Error_init(&eb);

    System_printf("Initializing master decoder %d\n", decoderIndex);
    System_flush();

    if (isMulticore) {
        TranscodeComponent_reserveCores(&decoderComponent->transcodeComponent, ALL_INSTANCES);
    }

    decoder = H264HighProfileDecoder_create(decoderConfiguration, &decoderComponent->transcodeComponent);

    if (isMulticore) {
        TranscodeComponent_releaseCores(&decoderComponent->transcodeComponent, ALL_INSTANCES);
    }

    if (decoder == NULL) {
        multicoreErrorToSet = ResultCode_Ok; // Do not set multi-core error code
        System_printf("Failed to create decoder\n");
        goto TASK_END;
    }

    System_printf("Created decoder %d %dx%d@%d\n", decoderIndex,
            decoderConfiguration->inputWidth,
            decoderConfiguration->inputHeight,
            decoderConfiguration->inputFrameRate);

    decoderHandle = decoder->handle;
    decoderStatus = decoder->status;
    inDecoderArguments = decoder->inArguments;
    outDecoderArguments = decoder->outArguments;
    decoderDynamicParams = decoder->dynamicParams;
    decoderFunctions = decoder->handle->fxns;

    System_printf("Configure decoder buffers\n");

    decoder->activate(decoder);
    result = decoderFunctions->control(decoderHandle, XDM_GETBUFINFO, decoderDynamicParams, decoderStatus);
    decoder->deactivate(decoder);

    if (result != IVIDDEC2_EOK) {
        System_printf("XDM_GETBUFINFO failed (%d)\n", result);
        goto TASK_END;
    }

    System_printf("minNumInBufs %d\nminNumOutBufs %d\n", decoderStatus->bufInfo.minNumInBufs, decoderStatus->bufInfo.minNumOutBufs);

    for (i = 0; i < decoderStatus->bufInfo.minNumInBufs; i++) {
        System_printf("minInBufSize[%d] %d\n", i, decoderStatus->bufInfo.minInBufSize[i]);
    }

    for (i = 0; i < decoderStatus->bufInfo.minNumOutBufs; i++) {
        System_printf("minOutBufSize[%d] %d\n", i, decoderStatus->bufInfo.minOutBufSize[i]);
    }

    if (decoderStatus->bufInfo.minNumInBufs != 1) {
        System_printf("Program is designed to work with single input buffer (%d input buffers required)\n",
                decoderStatus->bufInfo.minNumInBufs);
        goto TASK_END;
    }

    if (decoderStatus->bufInfo.minNumOutBufs != 3) {
        System_printf("Only 3 YUV planes supported\n");
        goto TASK_END;
    }

    inputDecoderBufferSize = decoderStatus->bufInfo.minInBufSize[0];
    inputDecoderBufferSize = (inputDecoderBufferSize / EDMA_SIZE_ALIGNMENT + 1) * EDMA_SIZE_ALIGNMENT;
    inputDecoderBuffer = Memory_alloc(ddr3Heap, inputDecoderBufferSize, L2_CACHE_LINE_SIZE, &eb);

    if (inputDecoderBuffer == NULL) {
        System_printf("Failed to allocate input decoder buffer\n");
        goto TASK_END;
    }

    inDecoderBufferDescriptor.numBufs = 1;
    inDecoderBufferDescriptor.descs[0].bufSize = inputDecoderBufferSize;
    inDecoderBufferDescriptor.descs[0].buf = (XDAS_Int8*) inputDecoderBuffer;

    decoderBufferManager = BufferDescriptorManager_create(512 * (maxEncodersCount + 1), decoderStatus->bufInfo.minNumOutBufs, decoderStatus->bufInfo.minOutBufSize);

    if (decoderBufferManager == NULL) {
        System_printf("Failed to create buffer manager for decoder\n");
        goto TASK_END;
    }

    inDecoderArguments->inputID = 0;

    System_printf("Master decoder %d initialized\n", decoderIndex);
    System_flush();

    while (1) {
        System_printf("----------\n");
        System_flush();

        Task_yield();

        if (inputBufferPointer >= inputBufferEndAddress) {
            System_printf("No more data. Test end.");
            System_flush();
            goto TASK_END;
        }

        if (isNewInputFrameNeeded) {
            if (bytesAvailable != inputDecoderBufferSize) {
                if ((inputBufferPointer < inputBufferEndAddress) && (bytesAvailable < inputDecoderBufferSize)) {
                    int bytesToRead = inputDecoderBufferSize - bytesAvailable;
                    if (bytesToRead > inputBufferEndAddress - inputBufferPointer) {
                        bytesToRead = inputBufferEndAddress - inputBufferPointer;
                    }

                    memcpy(inDecoderBufferDescriptor.descs[0].buf + bytesAvailable, inputBufferPointer, bytesToRead);

                    bytesAvailable += bytesToRead;
                    inputBufferPointer += bytesToRead;
                } else {
                    System_printf("No more data. Test end.");
                    System_flush();
                    goto TASK_END;
                }
            }
        }

        // Prepare input and output structures

        if (outDecoderBufferDescriptor == NULL) { // New output decoder buffer needed
            // ID of the new output buffer
            if (inDecoderArguments->inputID == INT_MAX) { // Handle integer overflow
                inDecoderArguments->inputID = 1;
            } else {
                inDecoderArguments->inputID++;
            }

            outDecoderBufferDescriptor = BufferDescriptorManager_getUnusedBufferDescriptor(decoderBufferManager, inDecoderArguments->inputID);

            if (outDecoderBufferDescriptor == NULL) {
                System_printf("Failed to create output buffer\n");
                goto TASK_END;
            }
        }

        inDecoderArguments->numBytes = bytesAvailable;

        if (isMulticore) {
            ResultCode errorCode = TranscodeComponent_waitErrorFromInstancesWithYield(&decoderComponent->transcodeComponent, decoder->groupId);
            if (errorCode != ResultCode_Ok) {
                multicoreErrorToSet = ResultCode_Ok; // Do not set multi-core error
                resultError = errorCode;
                System_printf("Error in encoder on another core\n");
                goto TASK_END;
            }

            // Share data with slave decoders

            gateKey = GateMP_enter(transcodeComponentGate);
            Cache_inv(decoderComponent, sizeof(*decoderComponent), Cache_Type_ALL, TRUE);
            decoderComponent->inBufferDescriptor = inDecoderBufferDescriptor.descs[0];
            decoderComponent->outBuffer_bufSizes[0] = outDecoderBufferDescriptor->bufSizes[0];
            decoderComponent->outBuffer_bufSizes[1] = outDecoderBufferDescriptor->bufSizes[1];
            decoderComponent->outBuffer_bufSizes[2] = outDecoderBufferDescriptor->bufSizes[2];
            decoderComponent->outBuffer_bufs[0] = outDecoderBufferDescriptor->bufs[0];
            decoderComponent->outBuffer_bufs[1] = outDecoderBufferDescriptor->bufs[1];
            decoderComponent->outBuffer_bufs[2] = outDecoderBufferDescriptor->bufs[2];
            memcpy(&decoderComponent->inArguments, inDecoderArguments, inDecoderArguments->size);
            Cache_wb(decoderComponent, sizeof(*decoderComponent), Cache_Type_ALL, TRUE);
            GateMP_leave(transcodeComponentGate, gateKey);

            // Wait for all decoder cores to be available
            TranscodeComponent_reserveCores(&decoderComponent->transcodeComponent, decoder->groupId);
        }

        // Decode

        Cache_wbInvAll();
        decoder->waitGroup(decoder);
        decoder->activate(decoder);
        System_printf("process begin\n");
        System_flush();
        unsigned long long startTicks = _itoll(TSCH, TSCL);
        result = decoderFunctions->process(decoderHandle, &inDecoderBufferDescriptor, outDecoderBufferDescriptor, inDecoderArguments, outDecoderArguments);
        unsigned long long endTicks = _itoll(TSCH, TSCL);
        System_printf("process end\n");
        System_flush();
        decoder->deactivate(decoder);

        Cache_wbInvAll();

        if (isMulticore) {
            decoder->waitGroup(decoder);
            TranscodeComponent_releaseCores(&decoderComponent->transcodeComponent, decoder->groupId);
        }

        if (result != IVIDDEC2_EOK) {
            decoder->waitGroup(decoder);
            decoder->activate(decoder);
            int getStatusResult = decoderFunctions->control(decoderHandle, XDM_GETSTATUS, decoderDynamicParams, decoderStatus);
            decoder->deactivate(decoder);

            System_printf("Decoder process error %d. ", result);

            if (getStatusResult != IVIDDEC2_EOK) {
                System_printf("Get status failed\n");
                resultError = ResultCode_ErrorNonFatal;
                multicoreErrorToSet = ResultCode_ErrorNonFatal;
                goto TASK_END;
            }

            int extendedError = decoderStatus->extendedError;

            System_printf("Extended error 0x%x - ", extendedError);

            if (XDM_ISFATALERROR(extendedError)) {
                System_printf("fatal error\n");
                resultError = ResultCode_ErrorNonFatal;
                multicoreErrorToSet = ResultCode_ErrorNonFatal;
                goto TASK_END;
            } else {
                System_printf("non-fatal error\n");
            }

            System_flush();
        } else {
            if (isMulticore) {
                decoder->waitGroup(decoder);
            }
        }

        System_printf("decode ticks=%u bytesConsumed=%d\n", (unsigned) (endTicks - startTicks), outDecoderArguments->bytesConsumed);
        System_flush();

        if (!outDecoderArguments->outBufsInUseFlag) {
            // Need to supply new output buffers for next "process" call
            outDecoderBufferDescriptor = NULL;
        }

        isNewInputFrameNeeded = (outDecoderArguments->bytesConsumed != 0);

        if (outDecoderArguments->bytesConsumed >= bytesAvailable) { // Decoder may consume more bytes than available
            bytesAvailable = 0;
        } else {
            bytesAvailable -= outDecoderArguments->bytesConsumed;

            memmove(inDecoderBufferDescriptor.descs[0].buf,
                inDecoderBufferDescriptor.descs[0].buf + outDecoderArguments->bytesConsumed,
                bytesAvailable);
        }

        // Output buffers that are ready

        i = 0;
        while (1) {
            int outputId = outDecoderArguments->outputID[i];

            if (outputId == 0) {
                break;
            }

            const IVIDEO1_BufDesc* displayBuffer = &outDecoderArguments->displayBufs[i];

            System_printf("outputID = %d\n", outputId);

            for (i = 0; i < 3; i++) {
                /*System_printf("Output size = %d\n", displayBuffer->bufDesc[i].bufSize);
                System_flush();*/
                memcpy(outputBufferPointer, displayBuffer->bufDesc[i].buf, displayBuffer->bufDesc[i].bufSize);
                outputBufferPointer += displayBuffer->bufDesc[i].bufSize;
            }

            i++;
        }

        // Release buffers not needed to decoder

        int* bufferIdToFreePointer = outDecoderArguments->freeBufID;

        while (*bufferIdToFreePointer != 0) {
            if (BufferDescriptorManager_releaseBufferDescriptor(decoderBufferManager, *bufferIdToFreePointer) != 0) {
                System_printf("Failed to release out decoder buffer\n");
                goto TASK_END;
            }

            System_printf("freeBufID %d\n", *bufferIdToFreePointer);

            bufferIdToFreePointer++;
        }

        System_flush();
    }

TASK_END:

    System_flush();

    if (multicoreErrorToSet != ResultCode_Ok) {
        TranscodeComponent_setErrorAndWaitInstancesWithYield(&decoderComponent->transcodeComponent, multicoreErrorToSet, decoder->groupId);
    }

    if (decoder != NULL) {
        if (isMulticore) {
            TranscodeComponent_reserveCores(&decoderComponent->transcodeComponent, ALL_INSTANCES);
        }

        decoder->delete(decoder);

        if (isMulticore) {
            TranscodeComponent_releaseCores(&decoderComponent->transcodeComponent, ALL_INSTANCES);
        }
    }

    if (inputDecoderBuffer != NULL) {
        Memory_free(ddr3Heap, inputDecoderBuffer, inputDecoderBufferSize);
    }

    if (decoderBufferManager != NULL) {
        BufferDescriptorManager_delete(decoderBufferManager);
    }

    System_flush();
    return resultError;
}
