#include <string.h>
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/family/c66/Cache.h>

#include "H264HighProfileDecoder.h"
#include "Heaps.h"
#include "Configuration.h"
#include "DecodeTaskCommon.h"
#include "Common.h"

extern cregister volatile unsigned int DNUM;
extern cregister volatile unsigned int TSCH;
extern cregister volatile unsigned int TSCL;

static ResultCode doSlaveDecoding(int decoderIndex, DecoderComponent* decoderComponent);

void decodeSlaveTask(UArg decoderIndex, UArg arg1)
{
    DecoderComponent* decoderComponent = NULL;

    System_printf("Get decoder component\n");
    System_flush();

    do {
        decoderComponent = (DecoderComponent*) TranscodeComponent_findById(Common_getDecoderAlgorithmIndex(decoderIndex));
        Task_sleep(1);
    } while (decoderComponent == NULL);

    while (1) {
        if (TranscodeComponent_waitErrorFromInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES) != ResultCode_Ok) {
            System_printf("Error in decoder on another core\n");
            goto TASK_END;
        }

        ResultCode error = doSlaveDecoding(decoderIndex, decoderComponent);

        switch (error) {
        case ResultCode_Error:
            System_printf("Fatal slave decoder task error. Terminating task.\n");
            goto TASK_END;

        case ResultCode_ErrorNonFatal:
            System_printf("Non fatal slave decoder task error. Restarting task.\n");
            break;

        default:
            System_printf("Unknown slave decoder task error. Terminating task.\n");
            goto TASK_END;
        }

        System_flush();

        // Wait master decoder
        TranscodeComponent_waitInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES);
        TranscodeComponent_waitInstancesWithYield(&decoderComponent->transcodeComponent, ALL_INSTANCES);
    }

TASK_END:

    System_printf("Slave decoder #%d task terminated\n", decoderIndex);
    System_flush();
}

static ResultCode doSlaveDecoding(int decoderIndex, DecoderComponent* decoderComponent)
{
    int result;
    ResultCode resultError = ResultCode_Error;
    ResultCode multicoreErrorToSet = ResultCode_Error;
    Error_Block eb;
    Decoder* decoder = NULL;
    const DspConfiguration* configuration = Configuration_get();
    const DspDecoderConfiguration* decoderConfiguration = &configuration->decoderConfigurations[decoderIndex];
    IVIDDEC2_Handle decoderHandle;
    IVIDDEC2_Fxns *decoderFunctions;
    IVIDDEC2_DynamicParams *decoderDynamicParams;
    IVIDDEC2_Status* decoderStatus;
    IVIDDEC2_OutArgs* outDecoderArguments;
    int isGroupMaster;

    // Shared data, synchronization, multi-core
    int algorithmIndex = Common_getDecoderAlgorithmIndex(decoderIndex);

    Error_init(&eb);

    System_printf("Initializing slave decoder %d\n", decoderIndex);
    System_flush();

    TranscodeComponent_waitCoreReservation(&decoderComponent->transcodeComponent, DNUM);

    decoder = H264HighProfileDecoder_create(decoderConfiguration, &decoderComponent->transcodeComponent);

    if (decoder == NULL) {
        System_printf("Failed to create decoder\n");
        multicoreErrorToSet = ResultCode_Ok; // Do not set multi-core error
        goto TASK_END;
    }

    decoderHandle = decoder->handle;
    decoderStatus = decoder->status;
    decoderDynamicParams = decoder->dynamicParams;
    outDecoderArguments = decoder->outArguments;
    decoderFunctions = decoder->handle->fxns;
    isGroupMaster = decoder->isGroupMaster;

    System_printf("Slave decoder %d initialized\n", decoderIndex);
    System_flush();

    while (1) {
        Task_yield();

        ResultCode errorCode = TranscodeComponent_waitErrorFromInstancesWithSleep(&decoderComponent->transcodeComponent, decoder->groupId);
        if (errorCode != ResultCode_Ok) {
            System_printf("Error in decoder on another core\n");
            multicoreErrorToSet = ResultCode_Ok; // Do not set multi-core error
            resultError = errorCode;
            goto TASK_END;
        }

        if (isGroupMaster) {
            TranscodeComponent_reserveCores(&decoderComponent->transcodeComponent, decoder->groupId);
        } else {
            TranscodeComponent_waitCoreReservation(&decoderComponent->transcodeComponent, DNUM);
        }

        Cache_wbInvAll();
        decoder->waitGroup(decoder);
        decoder->activate(decoder);
        result = decoderFunctions->process(decoderHandle, NULL, NULL, NULL, outDecoderArguments);
        decoder->deactivate(decoder);

        // Wait for other decoders (slaves and master)
        if (isGroupMaster) {
            decoder->waitGroup(decoder);
            TranscodeComponent_releaseCores(&decoderComponent->transcodeComponent, decoder->groupId);
        } else {
            decoder->waitGroup(decoder);
        }


        if (result != IVIDDEC2_EOK) {
            decoder->waitGroup(decoder);
            decoder->activate(decoder);
            int getStatusResult = decoderFunctions->control(decoderHandle, XDM_GETSTATUS, decoderDynamicParams, decoderStatus);
            decoder->deactivate(decoder);

            System_printf("Decoder process error %d. ", result);

            if (getStatusResult != IVIDDEC2_EOK) {
                System_printf("Get status failed\n");
                resultError = ResultCode_ErrorNonFatal;
                multicoreErrorToSet = ResultCode_ErrorNonFatal;
                goto TASK_END;
            }

            int extendedError = decoderStatus->extendedError;

            System_printf("Extended error 0x%x - ", extendedError);

            if (XDM_ISFATALERROR(extendedError)) {
                System_printf("fatal error\n");
                resultError = ResultCode_ErrorNonFatal;
                multicoreErrorToSet = ResultCode_ErrorNonFatal;
                goto TASK_END;
            } else {
                System_printf("non-fatal error\n");
            }

            System_flush();
        } else {
            decoder->waitGroup(decoder);
        }
    }

TASK_END:

    System_flush();

    if (multicoreErrorToSet != ResultCode_Ok) {
        TranscodeComponent_setErrorAndWaitInstancesWithYield(&decoderComponent->transcodeComponent, multicoreErrorToSet, decoder->groupId);
    }

    if (decoder != NULL) {
        TranscodeComponent_waitCoreReservation(&decoderComponent->transcodeComponent, DNUM);
        decoder->delete(decoder);
    }

    System_flush();
    return resultError;
}
