/**
 * Includes 4x or 2x H.264 HP decoder's code.
 */

#include <xdc/cfg/global.h>

/**
 * H.264 High Profile decoder that can run on 4 cores.
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <xdc/std.h>
#include <xdc/runtime/System.h>
#include <ti/sdo/fc/rman/rman.h>
#include <ti/sdo/codecs/h264hpvdec/ih264hpvdec.h>

#include "AlgorithmUtils.h"
#include "H264HighProfileDecoder.h"
#include "MulticoreApi.h"
#include "Configuration.h"
#include "Common.h"

extern IH264HPVDEC_Fxns H264HPVDEC_TI_IH264HPVDEC;
extern IRES_Fxns H264HPVDEC_TI_IRES;

extern cregister volatile unsigned int DNUM;

typedef struct {
    Decoder decoder;
    IH264HPVDEC_InArgs inArguments;
    IH264HPVDEC_OutArgs outArguments;
    IH264HPVDEC_Status status;
    IH264HPVDEC_Params params;
    IH264HPVDEC_DynamicParams dynamicParams;
    IVIDMC3_t multicoreParams;
    int isResourcesAssigned;
    int width;
    int height;
} H264HighProfileDecoder;

static void H264HighProfileDecoder_activate(Decoder* decoder)
{
    decoder->functions->ialg.algActivate((IALG_Handle) decoder->handle);
}

static void H264HighProfileDecoder_deactivate(Decoder* decoder)
{
    decoder->functions->ialg.algDeactivate((IALG_Handle) decoder->handle);
}

static void H264HighProfileDecoder_waitGroup(Decoder* decoder)
{
    TranscodeComponent_waitInstancesBlocking(decoder->component, decoder->groupId);
}

static void H264HighProfileDecoder_waitGroupDummy(Decoder* decoder)
{
}

static void H264HighProfileDecoder_syncAllSharedMemory(Decoder* decoder)
{
    TranscodeComponent_syncAllSharedMemory3(decoder->component);
}

static ResultCode H264HighProfileDecoder_initializeMulticoreApi(H264HighProfileDecoder *decoder, const DspDecoderConfiguration *decoderConfiguration, TranscodeComponent* component)
{
    int decoderTaskIndex = -1;
    int decodersCount = 0;

    int coreIndex;
    for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
        int coreId = decoderConfiguration->coreIds[coreIndex];

        if (coreId == -1) {
            break;
        }

        decodersCount++;

        if (coreId == DNUM) {
            if (decoderTaskIndex != -1) {
                System_printf("More then one DNUM in decoder configuration\n");
                return ResultCode_Error;
            }

            decoderTaskIndex = coreIndex;
        }
    }

    if (decoderTaskIndex == -1) {
        System_printf("Failed to find DNUM in decoder configuration\n");
        return ResultCode_Error;
    }

    if ((decodersCount == 3) || (decodersCount > 4)) {
        System_printf("Failed to initialize H.264 decoder. Only 1, 2 and 4 cores configurations supported.\n");
        return ResultCode_Error;
    }

    if (decodersCount == 4) {
        int32_t groupOfInstance[4] = { 0, 0, 1, 1 };

        if (decoderTaskIndex == 0) { // Master decoder updates group configuration
            if (TranscodeComponent_updateGroupsConfiguration(component, groupOfInstance) != ResultCode_Ok) {
                return ResultCode_Error;
            }
        }

        decoder->decoder.groupId = groupOfInstance[decoderTaskIndex];
    } else {
        decoder->decoder.groupId = 0;
    }

    decoder->multicoreParams.user_id = decoderTaskIndex;
    decoder->multicoreParams.num_users = decodersCount;

    if ((decoderTaskIndex == 0) || (decoderTaskIndex == 2)) {
        decoder->multicoreParams.task_ID = IVIDMC3_TASK_MASTER;
        decoder->decoder.isGroupMaster = 1;
    } else {
        decoder->multicoreParams.task_ID = IVIDMC3_TASK_LOCMASTER;
        decoder->decoder.isGroupMaster = 0;
    }

    System_printf("user_id=%d\n", decoder->multicoreParams.user_id);
    System_printf("num_users=%d\n", decoder->multicoreParams.num_users);
    System_printf("task_ID=%d\n", decoder->multicoreParams.task_ID);

    if (decodersCount > 1) { // Multi-core encoding
        decoder->multicoreParams.keyCreate = MulticoreApi3_keyCreate;
        decoder->multicoreParams.lockAcquire = TranscodeComponent_lockAcquire;
        decoder->multicoreParams.lockCheck = TranscodeComponent_lockCheck;
        decoder->multicoreParams.lockRelease = TranscodeComponent_lockRelease;
        decoder->multicoreParams.barrWait = TranscodeComponent_barrWait;
        decoder->decoder.waitGroup = H264HighProfileDecoder_waitGroup;
    } else { // Single core encoding
        decoder->multicoreParams.keyCreate = MulticoreApi3_keyCreate;
        decoder->multicoreParams.lockAcquire = MulticoreApi3_lockAcquireDummy;
        decoder->multicoreParams.lockCheck = MulticoreApi3_lockCheckDummy;
        decoder->multicoreParams.lockRelease = MulticoreApi3_lockReleaseDummy;
        decoder->multicoreParams.barrWait = MulticoreApi3_barrWaitDummy;
        decoder->decoder.waitGroup = H264HighProfileDecoder_waitGroupDummy;
    }

    decoder->multicoreParams.shmMap = TranscodeComponent_shmMap;
    decoder->multicoreParams.shmSync = TranscodeComponent_shmSync;
    decoder->multicoreParams.shmSyncWait = TranscodeComponent_shmSyncWait;

    decoder->params.coreTeam = calloc(decodersCount, sizeof(*decoder->params.coreTeam));

    if (decoder->params.coreTeam == NULL) {
        System_printf("Failed to allocate core team\n");
        return ResultCode_Error;
    }

    System_printf("core team:\n");
    int i;
    for (i = 0; i < decodersCount; i++) {
        System_printf("core #%d\n", decoderConfiguration->coreIds[i]);
        decoder->params.coreTeam[i] = decoderConfiguration->coreIds[i];
    }

    decoder->decoder.activate = H264HighProfileDecoder_activate;
    decoder->decoder.deactivate = H264HighProfileDecoder_deactivate;
    decoder->decoder.syncAllSharedMemory = H264HighProfileDecoder_syncAllSharedMemory;

    return ResultCode_Ok;
}

static void H264HighProfileDecoder_configure(H264HighProfileDecoder *decoder, int frameRate, int width, int height)
{
    decoder->params.mcViddecParams = &decoder->multicoreParams;
    decoder->params.viddecParams.size = sizeof(decoder->params);
    decoder->dynamicParams.viddecDynamicParams.size = sizeof(decoder->dynamicParams);

    decoder->params.viddecParams.maxFrameRate = frameRate * 1000;
    decoder->params.viddecParams.maxBitRate = 10000000; // TODO: should be configurable
    decoder->params.viddecParams.dataEndianness = XDM_BYTE;
    decoder->params.viddecParams.maxWidth = (width % 16 == 0) ? width : (width / 16) * 16 + 16; // Image width in pixels, must be multiple of 16
    decoder->params.viddecParams.maxHeight = (height % 16 == 0) ? height : (height / 16) * 16 + 16; // Image height in pixels, must be multiple of 16
    decoder->params.viddecParams.forceChromaFormat = XDM_YUV_420P; // 1 => XMI_YUV_420P, 3 => XMI_YUV_422IBE, 4 => XMI_YUV_422ILE
    decoder->params.displayDelay = 16;

    // TODO: remove this maxWidth and maxHeight adjustments when fixed in decoder

    if (decoder->params.viddecParams.maxWidth < 1280) {
        decoder->params.viddecParams.maxWidth = 1280;
    }

    if (decoder->params.viddecParams.maxHeight < 720) {
        decoder->params.viddecParams.maxHeight = 720;
    }

    System_printf("width=%d\n", decoder->params.viddecParams.maxWidth);
    System_printf("height=%d\n", decoder->params.viddecParams.maxHeight);

    decoder->dynamicParams.viddecDynamicParams.decodeHeader = XDM_DECODE_AU;
    decoder->dynamicParams.viddecDynamicParams.frameSkipMode = IVIDEO_NO_SKIP;
    decoder->dynamicParams.viddecDynamicParams.frameOrder = IVIDDEC2_DISPLAY_ORDER;
}

static void H264HighProfileDecoder_delete(Decoder *decoder)
{
    int result;
    H264HighProfileDecoder* h264Decoder = (H264HighProfileDecoder*) decoder;

    if (h264Decoder != NULL) {
        if (h264Decoder->isResourcesAssigned) {
            decoder->activate(decoder);
            result = RMAN_deactivateAllResources((IALG_Handle) decoder->handle, &H264HPVDEC_TI_IRES, 0);
            decoder->deactivate(decoder);
            if (result != IRES_OK) {
                System_printf("WARNING: Deactivate resources failed (%d)\n", result);
            }

            decoder->activate(decoder);
            result = RMAN_freeResources((IALG_Handle) decoder->handle, &H264HPVDEC_TI_IRES, 0);
            decoder->deactivate(decoder);
            if (result != IRES_OK) {
                System_printf("WARNING: Free resources failed (%d)\n", result);
            }
        }

        if (h264Decoder->decoder.handle != NULL) {
            ResultCode result = Algorithm_delete((IALG_Handle) decoder->handle);
            TranscodeComponent_deleteAlgorithmKeys(decoder->component);
            if (result != ResultCode_Ok) {
                System_printf("Failed to delete decoder\n");
            }
        }

        if (h264Decoder->params.coreTeam != NULL) {
            free(h264Decoder->params.coreTeam);
        }

        free(h264Decoder);
    }
}

Decoder* H264HighProfileDecoder_create(const DspDecoderConfiguration *decoderConfiguration, TranscodeComponent *component)
{
    int result;
    IVIDDEC2_Fxns* decoderFunctions;
    H264HighProfileDecoder* h264Decoder = malloc(sizeof(H264HighProfileDecoder));
    int numberOfDecoderCores = Configuration_calculateNumberOfCores(decoderConfiguration->coreIds);
    int isErrorFromOtherCore = 0;

    if (h264Decoder == NULL) {
        System_printf("Failed to allocate memory for decoder\n");
        goto CREATE_ERROR;
    }

    memset(h264Decoder, 0, sizeof(H264HighProfileDecoder));

    h264Decoder->decoder.inArguments = &h264Decoder->inArguments.viddecInArgs;
    h264Decoder->decoder.outArguments = &h264Decoder->outArguments.viddecOutArgs;
    h264Decoder->decoder.status = &h264Decoder->status.viddecStatus;
    h264Decoder->decoder.params = &h264Decoder->params.viddecParams;
    h264Decoder->decoder.dynamicParams = &h264Decoder->dynamicParams.viddecDynamicParams;
    h264Decoder->decoder.delete = H264HighProfileDecoder_delete;
    h264Decoder->decoder.component = component;

    h264Decoder->width = decoderConfiguration->inputWidth;
    h264Decoder->height = decoderConfiguration->inputHeight;

    h264Decoder->status.viddecStatus.size = sizeof(h264Decoder->status);
    h264Decoder->inArguments.viddecInArgs.size = sizeof(h264Decoder->inArguments);
    h264Decoder->outArguments.viddecOutArgs.size = sizeof(h264Decoder->outArguments);

    H264HighProfileDecoder_configure(h264Decoder, decoderConfiguration->inputFrameRate, decoderConfiguration->inputWidth, decoderConfiguration->inputHeight);

    if (H264HighProfileDecoder_initializeMulticoreApi(h264Decoder, decoderConfiguration, component) != ResultCode_Ok) {
        goto CREATE_ERROR;
    }

    if (TranscodeComponent_waitErrorFromInstancesBlocking(component, ALL_INSTANCES) != ResultCode_Ok) {
        System_printf("Error in decoder on another core\n");
        isErrorFromOtherCore = 1;
        goto CREATE_ERROR;
    }

    MulticoreApi_setActiveAlgorithm(component);
    h264Decoder->decoder.handle = (IVIDDEC2_Handle) Algorithm_create(
            &H264HPVDEC_TI_IH264HPVDEC.ividdec.ialg,
            (IALG_Params*) &h264Decoder->params,
            h264Decoder->decoder.memTabs);
    MulticoreApi_setActiveAlgorithm(NULL);

    if (h264Decoder->decoder.handle == NULL) {
        goto CREATE_ERROR;
    }

    if (TranscodeComponent_waitErrorFromInstancesBlocking(component, ALL_INSTANCES) != ResultCode_Ok) {
        System_printf("Error in decoder on another core\n");
        isErrorFromOtherCore = 1;
        goto CREATE_ERROR;
    }

    h264Decoder->decoder.functions = h264Decoder->decoder.handle->fxns;
    decoderFunctions = h264Decoder->decoder.functions;

    // Configure decoder resources

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = RMAN_assignResources((IALG_Handle) h264Decoder->decoder.handle, &H264HPVDEC_TI_IRES, 0);

    if (result != IRES_OK) {
        h264Decoder->decoder.deactivate(&h264Decoder->decoder);
        System_printf("Assign resources failed (%d)\n", result);
        goto CREATE_ERROR;
    }

    if (TranscodeComponent_waitErrorFromInstancesBlocking(component, ALL_INSTANCES) != ResultCode_Ok) {
        System_printf("Error in decoder on another core\n");
        isErrorFromOtherCore = 1;
        goto CREATE_ERROR;
    }

    h264Decoder->isResourcesAssigned = 1;

    result = RMAN_activateAllResources((IALG_Handle) h264Decoder->decoder.handle, &H264HPVDEC_TI_IRES, 0);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    //if (result != IRES_OK) {
        // TODO: WARNING: fix this
    //}

    // Configure decoder

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = decoderFunctions->control(h264Decoder->decoder.handle, XDM_RESET, &h264Decoder->dynamicParams.viddecDynamicParams, &h264Decoder->status.viddecStatus);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    if (result != IVIDDEC2_EOK) {
        System_printf("XDM_RESET failed (%d) with extended error (%d)\n", result, h264Decoder->status.viddecStatus.extendedError);
        goto CREATE_ERROR;
    }

    if (TranscodeComponent_waitErrorFromInstancesBlocking(component, ALL_INSTANCES) != ResultCode_Ok) {
        System_printf("Error in decoder on another core\n");
        isErrorFromOtherCore = 1;
        goto CREATE_ERROR;
    }

    // Set params

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = decoderFunctions->control(h264Decoder->decoder.handle, XDM_SETPARAMS, &h264Decoder->dynamicParams.viddecDynamicParams, &h264Decoder->status.viddecStatus);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    if (result != IVIDDEC2_EOK) {
        System_printf("XDM_SETPARAMS failed (%d) with extended error (%d)\n", result, h264Decoder->status.viddecStatus.extendedError);
        goto CREATE_ERROR;
    }

    if (TranscodeComponent_waitErrorFromInstancesBlocking(component, ALL_INSTANCES) != ResultCode_Ok) {
        System_printf("Error in decoder on another core\n");
        isErrorFromOtherCore = 1;
        goto CREATE_ERROR;
    }

    TranscodeComponent_waitInstancesBlocking(component, ALL_INSTANCES);
    h264Decoder->decoder.syncAllSharedMemory(&h264Decoder->decoder);
    Algorithm_synchronize(h264Decoder->decoder.memTabs);
    TranscodeComponent_waitInstancesBlocking(component, ALL_INSTANCES);

    return (Decoder*) h264Decoder;

CREATE_ERROR:

    if (!isErrorFromOtherCore) {
        TranscodeComponent_setErrorAndWaitInstancesBlocking(component, ResultCode_Error, ALL_INSTANCES);
    }

    H264HighProfileDecoder_delete((Decoder*) h264Decoder);
    return NULL;
}

