/**
 * Common functionality for multi-core transcode components: encoders, decoders, etc.
 */

#include "TranscodeComponent.h"

#include "Heaps.h"
#include "GlobalGates.h"

#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>

#include <ti/sysbios/family/c66/Cache.h>
#include <ti/sysbios/knl/Task.h>

#include <string.h>

#define SIUVIDMC3_CACHE_WBINVALL_THRESHOLD (358 * 128)

typedef struct TranscodeComponentListItem {
    struct TranscodeComponentListItem* next;
    TranscodeComponent component;
} TranscodeComponentListItem;

typedef struct {
    TranscodeComponentListItem* componentListHead;
    int coreUsedByMulticoreComponent[CORES_PER_DSP];
} TranscodeComponentsGlobal;

#pragma DATA_ALIGN(TRANSCODE_COMPONENTS_GLOBAL, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(TRANSCODE_COMPONENTS_GLOBAL, ".msmc")
static TranscodeComponentsGlobal TRANSCODE_COMPONENTS_GLOBAL;

void TranscodeComponent_initModule()
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    TRANSCODE_COMPONENTS_GLOBAL.componentListHead = NULL;

    int i;
    for (i = 0; i < CORES_PER_DSP; i++) {
        TRANSCODE_COMPONENTS_GLOBAL.coreUsedByMulticoreComponent[i] = -1;
    }

    Cache_wb(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);

    GateMP_leave(gate, gateKey);
}

static inline unsigned computeListItemSize(unsigned structureSize)
{
    return structureSize + sizeof(TranscodeComponentListItem) - sizeof(TranscodeComponent);
}

TranscodeComponent* TranscodeComponent_create(unsigned structureSize, TranscodeComponentType type, const int32_t* coreOfInstance, int componentId)
{
    int i;
    Error_Block eb;
    GateMP_Handle gate;
    IArg gateKey;
    TranscodeComponentListItem** currentItem;
    TranscodeComponentListItem* componentListItem = NULL;

    if (structureSize < sizeof(TranscodeComponent)) {
        goto CREATE_ERROR;
    }

    Error_init(&eb);

    componentListItem = Memory_alloc(msmcHeap, computeListItemSize(structureSize), L2_CACHE_LINE_SIZE, &eb);

    if (componentListItem == NULL) {
        goto CREATE_ERROR;
    }

    memset(componentListItem, 0, computeListItemSize(structureSize));

    componentListItem->next = NULL;
    componentListItem->component.structureSize = structureSize;
    componentListItem->component.id = componentId; // TODO: generate id
    componentListItem->component.type = type;
    componentListItem->component.error = ResultCode_Ok;

    for (i = 0; i < CORES_PER_DSP; i++) {
        if (coreOfInstance[i] == -1) {
            break;
        }

        componentListItem->component.numberOfInstances++;
    }

    componentListItem->component.groupSizes[0] = componentListItem->component.numberOfInstances;
    memcpy(componentListItem->component.coreOfInstance, coreOfInstance, sizeof(coreOfInstance[0]) * componentListItem->component.numberOfInstances);

    Cache_wb(componentListItem, computeListItemSize(structureSize), Cache_Type_ALL, TRUE);

    // Add component to components list
    gate = GlobalGates_getTranscodeComponentGate(); // TODO: change name to transcodeComponentGate
    gateKey = GateMP_enter(gate);
    Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);
    currentItem = &TRANSCODE_COMPONENTS_GLOBAL.componentListHead;

    while (*currentItem != NULL) {
        Cache_inv(*currentItem, sizeof(TranscodeComponentListItem), Cache_Type_ALL, TRUE);
        currentItem = &(*currentItem)->next;
    }

    *currentItem = componentListItem;

    Cache_wb(currentItem, sizeof(TranscodeComponentListItem*), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);

    return &componentListItem->component;

CREATE_ERROR:

    if (componentListItem != NULL) {
        Memory_free(msmcHeap, componentListItem, computeListItemSize(structureSize));
    }

    return NULL;
}

void TranscodeComponent_reset(TranscodeComponent* component)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);

    memset(&component->multicoreApiData, 0, sizeof(component->multicoreApiData));
    component->error = ResultCode_Ok;
    component->isInitialized = 0;

    Cache_wb(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

ResultCode TranscodeComponent_updateGroupsConfiguration(TranscodeComponent* component, const int32_t* groupOfInstance)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);

    memcpy(component->groupOfInstance, groupOfInstance, sizeof(groupOfInstance[0]) * component->numberOfInstances);

    memset(component->groupSizes, 0, sizeof(component->groupSizes));

    int i;
    for (i = 0; i < component->numberOfInstances; i++) {
        int32_t group = groupOfInstance[i];
        if ((group < 0) || (group >= MAX_ALGORITHM_GROUPS_PER_DSP)) {
            memset(component->groupSizes, 0, sizeof(component->groupSizes));
            Cache_wb(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
            GateMP_leave(gate, gateKey);
            return ResultCode_Error;
        }
        component->groupSizes[groupOfInstance[i]]++;
    }

    Cache_wb(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
    return ResultCode_Ok;
}

TranscodeComponent* TranscodeComponent_findById(int componentId)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);

    TranscodeComponentListItem* componentItem = TRANSCODE_COMPONENTS_GLOBAL.componentListHead;

    while (componentItem != NULL) {
        Cache_inv(componentItem, sizeof(TranscodeComponentListItem), Cache_Type_ALL, TRUE);
        if (componentItem->component.id == componentId) {
            break;
        }
        componentItem = componentItem->next;
    }

    if (componentItem != NULL) {
        Cache_inv(componentItem, computeListItemSize(componentItem->component.structureSize), Cache_Type_ALL, TRUE);
    }

    GateMP_leave(gate, gateKey);

    return (componentItem == NULL) ? NULL : &componentItem->component;
}

ResultCode TranscodeComponent_delete(int componentId)
{
    ResultCode result;
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);

    TranscodeComponentListItem** componentItem = &TRANSCODE_COMPONENTS_GLOBAL.componentListHead;

    while (*componentItem != NULL) {
        Cache_inv(*componentItem, sizeof(TranscodeComponentListItem), Cache_Type_ALL, TRUE);
        if ((*componentItem)->component.id == componentId) {
            break;
        }
        componentItem = &(*componentItem)->next;
    }

    if (*componentItem != NULL) {
        TranscodeComponentListItem* componentItemToDelete = *componentItem;
        *componentItem = componentItemToDelete->next;
        Cache_wb(componentItem, sizeof(TranscodeComponentListItem*), Cache_Type_ALL, TRUE);
        Memory_free(msmcHeap, componentItemToDelete, computeListItemSize(componentItemToDelete->component.structureSize));
        result = ResultCode_Ok;
    } else {
        result = ResultCode_Error;
    }

    GateMP_leave(gate, gateKey);
    return result;
}

static void TranscodeComponent_markCores(const int32_t* coreIds, const int32_t* groupIds, int marker, int32_t groupId)
{
    int i = 0;

    while (i < CORES_PER_DSP) {
        int coreId = coreIds[i];

        if (coreId == -1) {
            break;
        } else if ((groupId == ALL_INSTANCES) || (groupId == groupIds[i])) {
            TRANSCODE_COMPONENTS_GLOBAL.coreUsedByMulticoreComponent[coreId] = marker;
        }

        i++;
    }

    Cache_wb(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);
}

void TranscodeComponent_reserveCores(const TranscodeComponent* component, int32_t groupId)
{
    int i;
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();

    while (1) {
        IArg gateKey = GateMP_enter(gate);

        Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);

        int isRequiredCoresAvaliable = 1;

        for (i = 0; i < CORES_PER_DSP; i++) {
            int coreId = component->coreOfInstance[i];

            if (coreId == -1) {
                break;
            }

            if ((TRANSCODE_COMPONENTS_GLOBAL.coreUsedByMulticoreComponent[coreId] != -1) && ((groupId == ALL_INSTANCES) || (groupId == component->groupOfInstance[i]))) {
                isRequiredCoresAvaliable = 0;
                break;
            }
        }

        if (isRequiredCoresAvaliable) {
            TranscodeComponent_markCores(component->coreOfInstance, component->groupOfInstance, component->id, groupId);
            GateMP_leave(gate, gateKey);
            break;
        } else {
            GateMP_leave(gate, gateKey);
            Task_sleep(1);
        }
    }
}

void TranscodeComponent_releaseCores(const TranscodeComponent* component, int32_t groupId)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);
    TranscodeComponent_markCores(component->coreOfInstance, component->groupOfInstance, -1, groupId);

    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_waitCoreReservation(const TranscodeComponent* component, int coreId)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();

    while (1) {
        IArg gateKey = GateMP_enter(gate);

        Cache_inv(&TRANSCODE_COMPONENTS_GLOBAL, sizeof(TRANSCODE_COMPONENTS_GLOBAL), Cache_Type_ALL, TRUE);

        if (TRANSCODE_COMPONENTS_GLOBAL.coreUsedByMulticoreComponent[coreId] == component->id) {
            GateMP_leave(gate, gateKey);
            break;
        } else {
            GateMP_leave(gate, gateKey);
            Task_yield();
        }
    }
}

void TranscodeComponent_waitInstancesWithYield(TranscodeComponent* component, int32_t groupId)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey;
    int* barrierCounter;
    int numberOfInstances;

    if (groupId == ALL_INSTANCES) {
        barrierCounter = &component->componentBarrierCounter;
        numberOfInstances = component->numberOfInstances;
    } else {
        barrierCounter = &component->groupBarrierCounter[groupId];
        numberOfInstances = component->groupSizes[groupId];
    }

    // Wait while the barrier became available
    while (1) {
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

        if (*barrierCounter >= 0) {
            break;
        }
        Task_yield();
    }

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

    if (*barrierCounter == numberOfInstances - 1) { // Last task entered barrier
        *barrierCounter = -numberOfInstances + 1; // Leave barrier
    } else {
        (*barrierCounter)++;
        Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        while (1) {
            Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
            if (*barrierCounter < 0) {
                break;
            }
            Task_yield();
        }

        gateKey = GateMP_enter(gate);
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

        (*barrierCounter)++; // Leave barrier
    }

    Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_waitInstancesWithSleep(TranscodeComponent* component, int32_t groupId)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey;
    int* barrierCounter;
    int numberOfInstances;

    if (groupId == ALL_INSTANCES) {
        barrierCounter = &component->componentBarrierCounter;
        numberOfInstances = component->numberOfInstances;
    } else {
        barrierCounter = &component->groupBarrierCounter[groupId];
        numberOfInstances = component->groupSizes[groupId];
    }

    // Wait while the barrier became available
    while (1) {
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

        if (*barrierCounter >= 0) {
            break;
        }
        Task_sleep(1);
    }

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

    if (*barrierCounter == numberOfInstances - 1) { // Last task entered barrier
        *barrierCounter = -numberOfInstances + 1; // Leave barrier
    } else {
        (*barrierCounter)++;
        Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        while (1) {
            Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
            if (*barrierCounter < 0) {
                break;
            }
            Task_sleep(1);
        }

        gateKey = GateMP_enter(gate);
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

        (*barrierCounter)++; // Leave barrier
    }

    Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

static void TranscodeComponent_waitBarrierInternal(int *barrierCounter, int numberOfInstances)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey;

    // Wait while the barrier became available
    do {
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
    } while (*barrierCounter < 0);

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

    if (*barrierCounter == numberOfInstances - 1) { // Last task entered barrier
        *barrierCounter = -numberOfInstances + 1; // Leave barrier
    } else {
        (*barrierCounter)++;
        Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        do {
            Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        } while (*barrierCounter >= 0);

        gateKey = GateMP_enter(gate);
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);

        (*barrierCounter)++; // Leave barrier
    }

    Cache_wb(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_waitInstancesBlocking(TranscodeComponent* component, int32_t groupId)
{
    if (groupId == ALL_INSTANCES) {
        TranscodeComponent_waitBarrierInternal(&component->componentBarrierCounter, component->numberOfInstances);
    } else {
        TranscodeComponent_waitBarrierInternal(&component->groupBarrierCounter[groupId], component->groupSizes[groupId]);
    }

}

static void TranscodeComponent_setError(TranscodeComponent* component, ResultCode error)
{
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);

    component->error = error;

    Cache_wb(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_setErrorAndWaitInstancesBlocking(TranscodeComponent* component, ResultCode error, int32_t groupId)
{
    TranscodeComponent_setError(component, error);
    TranscodeComponent_waitInstancesBlocking(component, groupId);
    TranscodeComponent_waitInstancesBlocking(component, groupId);
}

void TranscodeComponent_setErrorAndWaitInstancesWithYield(TranscodeComponent* component, ResultCode error, int32_t groupId)
{
    TranscodeComponent_setError(component, error);
    TranscodeComponent_waitInstancesWithYield(component, groupId);
    TranscodeComponent_waitInstancesWithYield(component, groupId);
}

ResultCode TranscodeComponent_waitErrorFromInstancesWithSleep(TranscodeComponent* component, int32_t groupId)
{
    TranscodeComponent_waitInstancesWithSleep(component, groupId);

    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    ResultCode result = component->error;

    TranscodeComponent_waitInstancesWithYield(component, groupId);
    return result;
}

ResultCode TranscodeComponent_waitErrorFromInstancesWithYield(TranscodeComponent* component, int32_t groupId)
{
    TranscodeComponent_waitInstancesWithYield(component, groupId);

    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    ResultCode result = component->error;

    TranscodeComponent_waitInstancesWithYield(component, groupId);
    return result;
}

ResultCode TranscodeComponent_waitErrorFromInstancesBlocking(TranscodeComponent* component, int32_t groupId)
{
    TranscodeComponent_waitInstancesBlocking(component, groupId);

    Cache_inv(component, sizeof(TranscodeComponent), Cache_Type_ALL, TRUE);
    ResultCode result = component->error;

    TranscodeComponent_waitInstancesBlocking(component, groupId);
    return result;
}

static uint32_t computeFnv1Hash32(const uint8_t* string)
{
    static const uint32_t FNV32_PRIME = 16777619U;
    static const uint32_t FNV32_OFFSET_BASIS = 2166136261U;

    const uint8_t *buffer = (const uint8_t*) string;

    uint32_t hash = FNV32_OFFSET_BASIS;

    while (1) {
        uint8_t octet = *(buffer++);
        if (octet == 0) {
            return hash;
        }
        hash *= FNV32_PRIME;
        hash ^= octet;
    }
}

void TranscodeComponent_swbarr(TranscodeComponent* component, int32_t barrierId, int numberOfInstances)
{
    TranscodeComponent_waitBarrierInternal(&component->multicoreApiData.api1.barrierCounters[barrierId], numberOfInstances);
}

void* TranscodeComponent_shmmap(TranscodeComponent* component, int32_t shmem_key, int32_t shmem_size, uint32_t attr, int32_t alignment)
{
    Error_Block eb;
    void* result;
    SharedMemoryBlock* sharedMemoryBlock = &component->multicoreApiData.api1.sharedMemoryBlocks[shmem_key];

    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(sharedMemoryBlock, sizeof(*sharedMemoryBlock), Cache_Type_ALL, TRUE);

    if (sharedMemoryBlock->base != NULL) {
        result = sharedMemoryBlock->base;
        goto MAP_RESULT;
    }

    switch (attr) {
    case IVIDMC_SHMEM_ATTRS_SL2:
        sharedMemoryBlock->heap = msmcHeap;
        alignment = (alignment <= L2_CACHE_LINE_SIZE) ? L2_CACHE_LINE_SIZE : alignment;
        break;

    case IVIDMC_SHMEM_ATTRS_DDR_CACHED:
        sharedMemoryBlock->heap = ddr3Heap;
        alignment = (alignment <= L2_CACHE_LINE_SIZE) ? L2_CACHE_LINE_SIZE : alignment;
        break;

    case IVIDMC_SHMEM_ATTRS_DDR_UNCACHED:
        sharedMemoryBlock->heap = ddr3UncachedHeap;
        break;

    default:
        result = NULL;
        goto MAP_RESULT;
    }

    Error_init(&eb);
    sharedMemoryBlock->base = Memory_alloc(sharedMemoryBlock->heap, shmem_size, alignment, &eb);
    sharedMemoryBlock->size = shmem_size;

    result = sharedMemoryBlock->base;

    Cache_wb(sharedMemoryBlock, sizeof(*sharedMemoryBlock), Cache_Type_ALL, TRUE);

MAP_RESULT:
    GateMP_leave(gate, gateKey);
    return result;
}

void TranscodeComponent_shmmunmap(TranscodeComponent* component, int32_t shmem_key)
{
    SharedMemoryBlock* sharedMemoryBlock = &component->multicoreApiData.api1.sharedMemoryBlocks[shmem_key];
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(sharedMemoryBlock, sizeof(*sharedMemoryBlock), Cache_Type_ALL, TRUE);

    Memory_free(sharedMemoryBlock->heap, sharedMemoryBlock->base, sharedMemoryBlock->size);
    sharedMemoryBlock->base = NULL;

    Cache_wb(sharedMemoryBlock, sizeof(*sharedMemoryBlock), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_syncAllSharedMemory(TranscodeComponent* component)
{
    Cache_inv(component->multicoreApiData.api1.sharedMemoryBlocks, sizeof(component->multicoreApiData.api1.sharedMemoryBlocks), Cache_Type_ALL, TRUE);

    int i;
    for (i = 0; i < IVIDMC_SHMEM_NUM_KEYS; i++) {
        SharedMemoryBlock* sharedMemoryBlock = &component->multicoreApiData.api1.sharedMemoryBlocks[i];
        if (sharedMemoryBlock->base != NULL) {
            Cache_wbInv(sharedMemoryBlock->base, sharedMemoryBlock->size, Cache_Type_ALL, TRUE);
        }
    }
}

void TranscodeComponent_shmmap_sync(TranscodeComponent* component, int32_t shmem_key)
{
    if (shmem_key == IVIDMC_SHMEM_KEY_CLEAN) {
        TranscodeComponent_syncAllSharedMemory(component);
        return;
    }

    SharedMemoryBlock* sharedMemoryBlock = &component->multicoreApiData.api1.sharedMemoryBlocks[shmem_key];
    Cache_inv(sharedMemoryBlock, sizeof(*sharedMemoryBlock), Cache_Type_ALL, TRUE);
    Cache_wbInv(sharedMemoryBlock->base, sharedMemoryBlock->size, Cache_Type_ALL, TRUE);
}

static IHeap_Handle getHeapByType(IVIDMC3_SHMEM_ATTRS heapType)
{
    switch (heapType) {
    case IVIDMC3_SHMEM_ATTRS_DDR_CACHED:
        return ddr3Heap;

    case IVIDMC3_SHMEM_ATTRS_DDR_UNCACHED:
        return ddr3UncachedHeap;

    case IVIDMC3_SHMEM_ATTRS_SL2:
        return msmcHeap;

    default:
        return NULL;
    }
}

static ResultCode createShmemKey(Ividmc3Key* key, int32_t user_id, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg)
{
    Error_Block eb;

    IHeap_Handle heap = getHeapByType(cfg->u.shm.attr);

    if (heap == NULL) {
        return ResultCode_Error;
    }

    Error_init(&eb);

    key->data.sharedMemory.base = Memory_alloc(heap, cfg->u.shm.size, cfg->u.shm.alignment, &eb);

    if (key->data.sharedMemory.base == NULL) {
        return ResultCode_Error;
    }

    key->data.sharedMemory.heapType = cfg->u.shm.attr;
    key->data.sharedMemory.size = cfg->u.shm.size;

    return ResultCode_Ok;
}

static ResultCode createSwbarrKey(Ividmc3Key* key, int32_t user_id, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg)
{
    key->data.barrier.counter = 0;
    key->data.barrier.usersCount = num_users;
    return ResultCode_Ok;
}

static ResultCode createLockKey(Ividmc3Key* key, int32_t user_id, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg)
{
    key->data.lock.userId = -1;
    return ResultCode_Ok;
}

Ividmc3Key* TranscodeComponent_keyCreate(TranscodeComponent* component, uint8_t* name, int32_t user_id, IVIDMC3_KEY_SPACE_e key_space, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg)
{
    Ividmc3Key* resultKey = NULL;
    uint32_t nameHash = computeFnv1Hash32(name);
    MulticoreApi3Data *multicoreData = &component->multicoreApiData.api3;

    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    // Find key by name hash

    Cache_inv(multicoreData, sizeof(*multicoreData), Cache_Type_ALL, TRUE);

    int i;
    Ividmc3Key* key;
    for (i = 0; i < multicoreData->keysCount; i++) {
        key = &multicoreData->keys[i];

        if ((key->nameHash == nameHash) && (key->keySpace == key_space)) { // Key found
            resultKey = key;
            goto KEY_CREATE_END;
        }
    }

    // Key not found - create new one

    if (multicoreData->keysCount >= MAX_IVIDMC3_KEYS) {
        goto KEY_CREATE_END; // No available keys
    }

    key = &multicoreData->keys[multicoreData->keysCount];
    key->nameHash = nameHash;
    key->keySpace = key_space;

    ResultCode (*createKeyFunction)(Ividmc3Key* key, int32_t user_id, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg);

    switch (key_space) {
    case IVIDMC3_SHMEM_SPACE:
        createKeyFunction = createShmemKey;
        break;

    case IVIDMC3_SWBARR_SPACE:
        createKeyFunction = createSwbarrKey;
        break;

    case IVIDMC3_LOCK_SPACE:
        createKeyFunction = createLockKey;
        break;

    default:
        goto KEY_CREATE_END; // Key type not supported
    }

    if (createKeyFunction(key, user_id, num_users, user_ids, cfg) != ResultCode_Ok) {
        goto KEY_CREATE_END;
    }

    multicoreData->keysCount++;
    resultKey = key;

    Cache_wb(multicoreData, sizeof(*multicoreData), Cache_Type_ALL, TRUE);

KEY_CREATE_END:
    GateMP_leave(gate, gateKey);
    return resultKey;
}

static void deleteShmemKey(Ividmc3Key* key)
{
    Memory_free(getHeapByType(key->data.sharedMemory.heapType), key->data.sharedMemory.base, key->data.sharedMemory.size);
}

void TranscodeComponent_deleteAlgorithmKeys(TranscodeComponent* component)
{
    MulticoreApi3Data *multicoreData = &component->multicoreApiData.api3;

    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(multicoreData, sizeof(*multicoreData), Cache_Type_ALL, TRUE);

    int i;
    Ividmc3Key* key;
    for (i = 0; i < multicoreData->keysCount; i++) {
        key = &multicoreData->keys[i];
        if (key->keySpace == IVIDMC3_SHMEM_SPACE) {
            deleteShmemKey(key);
        }
    }

    multicoreData->keysCount = 0;

    Cache_wb(multicoreData, sizeof(*multicoreData), Cache_Type_ALL, TRUE);

    GateMP_leave(gate, gateKey);
}

void TranscodeComponent_syncAllSharedMemory3(TranscodeComponent* component)
{
    MulticoreApi3Data *multicoreData = &component->multicoreApiData.api3;

    int i;
    Ividmc3Key* key;
    for (i = 0; i < multicoreData->keysCount; i++) {
        if (key->keySpace == IVIDMC3_SHMEM_SPACE) {
            Cache_wbInv(key->data.sharedMemory.base, key->data.sharedMemory.size, Cache_Type_ALL, TRUE);
        }
    }
}

int32_t TranscodeComponent_lockAcquire(int32_t user_id, void* lockHandle)
{
    LockKey* key = &((Ividmc3Key*)lockHandle)->data.lock;

    while (1) {
        GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
        IArg gateKey = GateMP_enter(gate);

        Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

        if (key->userId == -1) { // Lock is released
            key->userId = user_id;
            Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
            GateMP_leave(gate, gateKey);
            return IVIDMC3_LOCK_ERROR_NONE;
        } else if (key->userId == user_id) {
            GateMP_leave(gate, gateKey);
            return IVIDMC3_LOCK_ERROR_DBL_LOCK;
        }

        GateMP_leave(gate, gateKey);

        // Wait until lock is released
        do {
            Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);
        } while (key->userId == -1);
    }
}

int32_t TranscodeComponent_lockRelease(int32_t user_id, void* lockHandle)
{
    LockKey* key = &((Ividmc3Key*)lockHandle)->data.lock;
    int32_t result;
    GateMP_Handle gate = GlobalGates_getTranscodeComponentGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

    if (key->userId == user_id) {
        result = IVIDMC3_LOCK_ERROR_NONE;
        key->userId = -1;
        Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
    } else if (key->userId == -1) {
        result = IVIDMC3_LOCK_ERROR_DBL_UNLOCK;
    } else {
        result = IVIDMC3_LOCK_ERROR_USER_ID;
    }

    GateMP_leave(gate, gateKey);
    return result;
}

int32_t TranscodeComponent_lockCheck(int32_t user_id, void* lockHandle)
{
    LockKey* key = &((Ividmc3Key*)lockHandle)->data.lock;
    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);
    return (key->userId == -1) ? IVIDMC3_LOCK_ERROR_NONE : IVIDMC3_LOCK_IN_USE;
}

int32_t TranscodeComponent_barrWait(int32_t user_id, void* barrHandle)
{
    BarrierKey* key = &((Ividmc3Key*)barrHandle)->data.barrier;
    TranscodeComponent_waitBarrierInternal(&key->counter, key->usersCount);
    return IVIDMC3_BARR_ERROR_NONE;
}

int32_t* TranscodeComponent_shmMap(int32_t user_id, void *shmemHandle)
{
    return ((Ividmc3Key*)shmemHandle)->data.sharedMemory.base;
}

int32_t TranscodeComponent_shmSync(int32_t user_id, void *shmemHandle, int32_t* shmem_base, int32_t shmem_size, uint32_t shmem_sync_attribs)
{
    if (shmem_sync_attribs & IVIDMC3_SYNC_ATTRIBS_LOC_WRITEBACK) {
        if (shmem_size >= SIUVIDMC3_CACHE_WBINVALL_THRESHOLD) {
            Cache_wbInvAll();
        } else {
            Cache_wbInv(shmem_base, shmem_size, Cache_Type_ALL, TRUE);
        }
    } else if (shmem_sync_attribs & IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE) {
        if (shmem_size >= SIUVIDMC3_CACHE_WBINVALL_THRESHOLD) {
            Cache_wbInvAll();
        } else {
            Cache_inv(shmem_base, shmem_size, Cache_Type_ALL, TRUE);
        }
    }

    return 0;
}

int32_t TranscodeComponent_shmSyncWait(int32_t user_id, void *shmemHandle, int32_t* shmem_transid)
{
    return 0;
}
