/**
 * Common functionality for multi-core transcode components: encoders, decoders, etc.
 */

#pragma once

#include <xdc/runtime/IHeap.h>
#include <ti/xdais/xdas.h>

#include "Common.h"
#include "ividmc/ividmc.h"
#include "ividmc/ividmc3.h"

#define MAX_IVIDMC3_KEYS (24)
#define MAX_ALGORITHM_GROUPS_PER_DSP (4)
#define ALL_INSTANCES (-1)

typedef enum {
    TranscodeComponentType_Encoder,
    TranscodeComponentType_Decoder
} TranscodeComponentType;

typedef struct {
    void *base;
    int32_t size;
    IHeap_Handle heap;
} SharedMemoryBlock;

typedef struct {
    int barrierCounters[IVIDMC_NUM_BARRIERS];
    SharedMemoryBlock sharedMemoryBlocks[IVIDMC_SHMEM_NUM_KEYS];
} MulticoreApi1Data;

typedef struct {
    int32_t userId;
} LockKey;

typedef struct {
    int usersCount;
    int counter;
} BarrierKey;

typedef struct {
    void* base;
    IVIDMC3_SHMEM_ATTRS heapType;
    int32_t size;
} SharedMemoryKey;

typedef union {
    LockKey lock;
    BarrierKey barrier;
    SharedMemoryKey sharedMemory;
} KeyData;

typedef struct {
    int nameHash;
    IVIDMC3_KEY_SPACE_e keySpace;
    KeyData data;
} Ividmc3Key;

typedef struct {
    Ividmc3Key keys[MAX_IVIDMC3_KEYS];
    int keysCount;
} MulticoreApi3Data;

typedef union {
    MulticoreApi1Data api1;
    MulticoreApi3Data api3;
} MulticoreApiData;

typedef struct {
    // TODO: name
    unsigned structureSize;
    int id;
    int isInitialized;
    TranscodeComponentType type;
    int32_t coreOfInstance[CORES_PER_DSP];
    int32_t groupOfInstance[CORES_PER_DSP];
    int numberOfInstances;
    ResultCode error;
    int groupBarrierCounter[MAX_ALGORITHM_GROUPS_PER_DSP];
    int groupSizes[MAX_ALGORITHM_GROUPS_PER_DSP];
    int componentBarrierCounter;
    MulticoreApiData multicoreApiData;
} TranscodeComponent;

extern void TranscodeComponent_initModule();

extern TranscodeComponent* TranscodeComponent_create(unsigned structureSize, TranscodeComponentType type, const int32_t* coreOfInstance, int id);
extern ResultCode TranscodeComponent_updateGroupsConfiguration(TranscodeComponent* component, const int32_t* groupOfInstance);
extern void TranscodeComponent_reset(TranscodeComponent* component);
extern TranscodeComponent* TranscodeComponent_findById(int id);
extern ResultCode TranscodeComponent_delete(int id);
extern void TranscodeComponent_reserveCores(const TranscodeComponent* component, int32_t groupId);
extern void TranscodeComponent_releaseCores(const TranscodeComponent* component, int32_t groupId);
extern void TranscodeComponent_waitCoreReservation(const TranscodeComponent* component, int coreId);

extern void TranscodeComponent_waitInstancesWithYield(TranscodeComponent* component, int32_t groupId);
extern void TranscodeComponent_waitInstancesWithSleep(TranscodeComponent* component, int32_t groupId);
extern void TranscodeComponent_waitInstancesBlocking(TranscodeComponent* component, int32_t groupId);

extern void TranscodeComponent_setErrorAndWaitInstancesBlocking(TranscodeComponent* component, ResultCode error, int32_t groupId);
extern void TranscodeComponent_setErrorAndWaitInstancesWithYield(TranscodeComponent* component, ResultCode error, int32_t groupId);
extern ResultCode TranscodeComponent_waitErrorFromInstancesWithSleep(TranscodeComponent* component, int32_t groupId);
extern ResultCode TranscodeComponent_waitErrorFromInstancesWithYield(TranscodeComponent* component, int32_t groupId);
extern ResultCode TranscodeComponent_waitErrorFromInstancesBlocking(TranscodeComponent* component, int32_t groupId);

// IVIDMC API

extern void TranscodeComponent_swbarr(TranscodeComponent* component, int32_t barrierId, int numberOfInstances);
extern void* TranscodeComponent_shmmap(TranscodeComponent* component, int32_t shmem_key, int32_t shmem_size, uint32_t attr, int32_t alignment);
extern void TranscodeComponent_shmmunmap(TranscodeComponent* component, int32_t shmem_key);
extern void TranscodeComponent_syncAllSharedMemory(TranscodeComponent* component);
extern void TranscodeComponent_shmmap_sync(TranscodeComponent* component, int32_t shmem_key);

// IVIDMC3 API

extern Ividmc3Key* TranscodeComponent_keyCreate(TranscodeComponent* component, uint8_t* name, int32_t user_id, IVIDMC3_KEY_SPACE_e key_space, int32_t num_users, int32_t* user_ids, IVIDMC3_KeyCfg_t* cfg);
extern void TranscodeComponent_deleteAlgorithmKeys(TranscodeComponent* component);
extern void TranscodeComponent_syncAllSharedMemory3(TranscodeComponent* component);
extern int32_t TranscodeComponent_lockAcquire(int32_t user_id, void* lockHandle);
extern int32_t TranscodeComponent_lockRelease(int32_t user_id, void* lockHandle);
extern int32_t TranscodeComponent_lockCheck(int32_t user_id, void* lockHandle);
extern int32_t TranscodeComponent_barrWait(int32_t user_id, void* barrHandle);
extern int32_t* TranscodeComponent_shmMap(int32_t user_id, void* shmemHandle);
extern int32_t TranscodeComponent_shmSync(int32_t user_id, void* shmemHandle, int32_t* shmem_base, int32_t shmem_size, uint32_t shmem_sync_attribs);
extern int32_t TranscodeComponent_shmSyncWait(int32_t user_id, void* shmemHandle, int32_t* shmem_transid);
