/* FILE PURPOSE:  Video Codec Multicore Interface Header
******************************************************************************
* FILE NAME:   ividmc3.h
*
* DESCRIPTION: Video Codec Multicore Interface Header
*
* TABS: NONE
*
*
* (C) Copyright 2012, Texas Instruments Incorporated
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  @author   Vivek Chengalvala, Djordje Senicic
*/


#ifndef _IVIDMC3_H_  
#define _IVIDMC3_H_

#include <xdc/std.h>
#include <ti/xdais/xdas.h>

/*---------------------------------------*/
/* Shared memory attributes              */
/*---------------------------------------*/
#define IVIDMC3_SHMEM_ATTRS_DDR_CACHED   (0)
#define IVIDMC3_SHMEM_ATTRS_DDR_UNCACHED (1)
#define IVIDMC3_SHMEM_ATTRS_SL2          (2)
/* Remote global memory buffer is mirrored in local DDR */
#define IVIDMC3_SHMEM_ATTRS_RMT_UNCACHED_LOC_DDR (3)
/* Remote global memory buffer is mirrored in local SL2 */
#define IVIDMC3_SHMEM_ATTRS_RMT_UNCACHED_LOC_SL2 (4)
#define IVIDMC3_SHMEM_ATTRS_RMT_CACHED_LOC_DDR   (5)
#define IVIDMC3_SHMEM_ATTRS_RMT_CACHED_LOC_SL2   (6)
#define IVIDMC3_SHMEM_ATTRS_RMT_UNCACHED_NO_LOC  (7) /* No sync support for this
                                                       shmem */
/*----------------------------------------------------------------------------*/
/* Maximum number of cores running concurrently                               */
/*----------------------------------------------------------------------------*/
#define IVIDMC3_TI_MAXCORES 64

typedef enum {
  IVIDMC3_TASK_MASTER         = 0,   /** Global master                       */
  IVIDMC3_TASK_LOCMASTER      = 1,   /** Chip-local master                   */
  IVIDMC3_TASK_SLAVE          = 2    /** Slave cores                         */
} IVIDMC3_TASK_e;

enum {
  IVIDMC3_BARR_ERROR_NONE        =  0,  /** Function call was a success.      */

  IVIDMC3_BARR_ERROR_HANDLE      = -1,  /** Invalid handle passed to the 
                                      *  funciton.                            */

  IVIDMC3_BARR_ERROR_USER_ID     = -2,  /** Invalid user ID passed to the 
                                      *  funciton.                            */

  IVIDMC3_BARR_ERROR_CORRUPTION  = -3   /** Instance data shows that the user ID
                                      *  is already waiting at the barrier 
                                      *  even though the function had just 
                                      *  been entered. This can commonly occur 
                                      *  when the same barrier is used 
                                      *  consecutively.                       */
};

enum {
  IVIDMC3_LOCK_IN_USE            =  1,  /** Sync lock in use                  */
  IVIDMC3_LOCK_ERROR_NONE        =  0,  /** Function was a success.           */

  IVIDMC3_LOCK_ERROR_HANDLE      = -1,  /** Invalid handle was passed to the 
                                         *  funciton.                         */

  IVIDMC3_LOCK_ERROR_USER_ID     = -2,  /** Invalid user ID was passed to the 
                                         *  function.                         */

  IVIDMC3_LOCK_ERROR_DBL_LOCK    = -3,  /** Lock is attempted to be acquired by 
                                         *  a user ID that already holds the 
                                         *  lock.                             */

  IVIDMC3_LOCK_ERROR_DBL_UNLOCK  = -4   /** User ID attempted to release a lock 
                                         *  when it did not hold the lock.    */
};

typedef XDAS_UInt32 IVIDMC3_SHMEM_ATTRS;

/** Address spaces **/
typedef enum {
  IVIDMC3_SHMEM_SPACE  = 0,  /* Shmem keys */
  IVIDMC3_SWBARR_SPACE,      /* Software barrier */
  IVIDMC3_LOCK_SPACE,        /* Spinlocks */
  IVIDMC3_MAILBOX_SPACE      /* Mailboxs  */
} IVIDMC3_KEY_SPACE_e;


/* DIRECTION: Indicates from the local perspective whether the mailbox is used 
 * to send messages or Receive messages
 */
#define IVIDMC3_MAILBOX_DIRECTION_RECEIVE  0
#define IVIDMC3_MAILBOX_DIRECTION_SEND     1

typedef struct {
  XDAS_Int32 direction;   /* Use IVIDMC3_MAILBOX_DIRECTION_RECEIVE or _SEND to
                             define Tx or Rx endpoint of Mailbox */
  XDAS_Int32 max_mails;   /* # of mail slots */
  XDAS_Int32 mail_size;   /* max size of single mail */
} IVIDMC3_MailBoxConfig_t;

typedef struct {
  XDAS_Int32           size;   /* Size of shared memory block */
  IVIDMC3_SHMEM_ATTRS  attr;         /* Attributes - see _SHMEM_ATTRS */
  XDAS_Int32           alignment;    /* alignment requirement */
} IVIDMC3_ShMemCfg_t;


/* Data structure with additional details for IPC resource creation */
typedef struct {
  union {
    IVIDMC3_ShMemCfg_t       shm;
    IVIDMC3_MailBoxConfig_t  mailBox;
  }u;
} IVIDMC3_KeyCfg_t;


/* Indication that this is unused core. DNUM can not be equal to this value */
#define IVIDMC3_CORE_NOT_USED 0xFF

/* Indication to writeback invalidate entire cache */
#define IVIDMC3_SHMEMKEY_INV_ALL 0xFFFF

/* Bit field specifing synchronization attributes:      */
typedef XDAS_UInt32 IVIDMC3_SYNC_ATTRIBS;
#define IVIDMC3_SYNC_ATTRIBS_NONBLOCKING 0x0010

/** The below bits are to be OR'd together to tell the SYNC API the operation to
 *  perform on the buffer
 *  Current valid combinations are the follow:
 *  -  ( IVIDMC3_SYNC_ATTRIBS_LOC_WRITEBACK  | 
 *       IVIDMC3_SYNC_ATTRIBS_RMT_WRITEBACK  ) ==> Data in cache is valid. 
 *       Update local and remote buffers with cached data
 *  -  ( IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE | 
 *       IVIDMC3_SYNC_ATTRIBS_RMT_INVALIDATE ) ==> Data in remote buffer is 
 *       valid. Update local buffers with the remote and invalidate cache
 *  -  ( IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE | 
 *       IVIDMC3_SYNC_ATTRIBS_RMT_WRITEBACK  ) ==> Data is valid in local buffer
 *       (e.g. by DMA). Invalidate cache and update remote buffers
 *  -  ( IVIDMC3_SYNC_ATTRIBS_LOC_WRITEBACK  ) ==> Write back cache to local 
 *       buffer, but do not copy to remote buffer
 *  -  ( IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE ) ==> Invalidate cache of local 
 *       buffer, but do not copy contents of remote buffer
 */
#define IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE    0x0001   /** This will cause a 
                              cache invalidate on the local buffer */
#define IVIDMC3_SYNC_ATTRIBS_LOC_WRITEBACK     0x0002   /** This will cause a 
                              cache writeback on the local buffer */
#define IVIDMC3_SYNC_ATTRIBS_RMT_INVALIDATE    0x0004   /** This will cause the 
                              remote buffer to be copied to the local buffer */
#define IVIDMC3_SYNC_ATTRIBS_RMT_WRITEBACK     0x0008   /** This will cause the 
                              local buffer to be copied to the remote buffer */
/**
 *******************************************************************************
 *  @struct IVIDMC3_t
 *  @brief  This structure contains Multicore API's and shared variables 
 *          for use in multicore DSP codec
 *******************************************************************************
*/
typedef struct IVIDMC3_s {
/**
 *  @brief Key assignment for Barrier, Shmem, Locks and Mailboxes.
 *         Barrier, Shmem and Locks resources are also initialized
 *  @param[in]   name         Null terminated string (globally known)
 *  @param[in]   user_id      Core ID or generic user id
 *  @param[in]   key_space    Resource identifier (one out of four)
 *  @param[in]   num_users    Number of users for this shmem region. Also, 
 *                            number of ids in next array
 *  @param[in]   *user_ids    List of user ids using this shmem region. 
 *                            E.g. 0, 1, 2, 3, .., 
 *  @return      sync resource handle
 */
  XDAS_Void *(*keyCreate) (XDAS_UInt8 *name, XDAS_Int32 user_id, 
                           IVIDMC3_KEY_SPACE_e key_space, XDAS_Int32 num_users,
                           XDAS_Int32 *user_ids, IVIDMC3_KeyCfg_t *cfg);

/**
 *  @brief Function call which will block until all other users have also 
 *         called it. NOTE: Limitation: This cannot be called consecutively with
 *          the same barrier instance due to race conditions that can cause
 *          corruption
 *  @param[in]   barrHandle   Unique barrier handle
 *  @param[in]   user_id      ID of the user entering the barrier
 *  @return      0 if barrier is reached by all users, or error code
 */
  XDAS_Int32 (*barrWait)(XDAS_Int32 user_id, XDAS_Void *barrHandle);


/*
 * @brief Dynamically allocate shared memory region identified with handle 
 * "shmemHandle".
 *        If region with same "shmemHandle" is already allocated, its base
 *        address is provided
 * @param[in] user_id          Core ID or more generic user_id
 * @param[in] shmemHandle      Unique shmem handle
 * @param[in] shmem_size       Size of shmem region 
 * @param[in] attr             Attribute provided here indicates shared memory 
 *                             region placement: SL2 or DDR
 * @param[in] alignment        Meaning of alignment is identical to ialg/xdais 
 *                             definition. Having alignment in the API allows
 *                             the Codec to request a shared buffer with 
 *                             required alignment.
 * @return                     NULL on failure, valid ptr in local address 
 *                             space, on success
 */
  XDAS_Int32 *(*shmMap) (XDAS_Int32 user_id, XDAS_Void *shmemHandle);


/*
 * @brief Synchronize shared memory region:  (blocking or non-blocking call).
 * @param[in] user_id            Core ID or more generic user_id
 * @param[in] shmemHandle        Unique shmem handle
 * @param[in] shmem_base         Shared memory base - start of region to be 
 *                               synchronized
 * @param[in] shmem_size         Size of region to be synchronized
 * @param[in] shmem_sync_attribs Synchronization attributes
 * @return                       Return shmem_transID or error. For succesfull 
 *                               blocking call, returns 0
 */
#define IVIDMC3_SHMEM_SYNC_TRANSID_OK    0
#define IVIDMC3_SHMEM_SYNC_TRANSID_WAIT_FIRST  1
#define IVIDMC3_SHMEM_SYNC_TRANSID_WAIT_LAST   0x7FFFFFFF
#define IVIDMC3_SHMEM_SYNC_TRANSID_ERROR (-1)
  XDAS_Int32 (*shmSync)(XDAS_Int32 user_id, XDAS_Void *shmemHandle, 
               XDAS_Int32 *shmem_base, XDAS_Int32 shmem_size, 
               IVIDMC3_SYNC_ATTRIBS shmem_sync_attribs);

/*
 * @brief Wait for synchronization operation to finish (after _sync_rmt 
 *                                                      non-blocking call)
 * @param[in] user_id           Core ID or more generic user_id
 * @param[in] shmemHandle       shmem handle
 * @param[in] *shmem_transid    Transaction id of region to be synchronized     
 */
  XDAS_Int32 (*shmSyncWait)(XDAS_Int32 user_id, XDAS_Void *shmemHandle, 
               XDAS_Int32 *shmem_transid);

/* @brief Critical region lock acquire (multi-core environment)
 * If spinlock has been succesfully acquired, this function returns 1
 * If same spinlock has been already acquired by another participant,
 * this function returns 0
 * @param[in] lockHandle
 * @return 
*/
  XDAS_Int32  (*lockAcquire)  (XDAS_Int32 user_id, XDAS_Void *lockHandle);

/* @brief Critical region lock release (multi-core environment)
 * @param[in] lockHandle
 * @return 
*/
  XDAS_Int32  (*lockRelease)  (XDAS_Int32 user_id, XDAS_Void *lockHandle);

/* @brief Critical region lock release (multi-core environment)
 * @param[in] lockHandle
 * @return 
*/
  XDAS_Int32  (*lockCheck)    (XDAS_Int32 user_id, XDAS_Void *lockHandle);


/**
 *  @brief Function mailBox_open() Opens a mailBox; This is a blocking call, 
 *                                 wait till the remote is ready
 *  @param[in]     mailBoxHandle  mailBox Handle
 *  @retval        0 for success, -1 for failure
 *  @pre
 *  @post
 */
  XDAS_Int32 (*mailBoxOpen)(XDAS_Void *mailBoxHandle);

#define IVIDMC3_MAILBOX_ERR_FAIL          -1
#define IVIDMC3_MAILBOX_ERR_MAIL_BOX_FULL -2
#define IVIDMC3_MAILB0X_ERR_EMPTY         -3
#define IVIDMC3_MAILBOX_READ_ERROR        -4

/**
 *  @brief Function mailBoxWrite() Writes into a mailBox to deliver to remote 
                                   : Non blocking call
 *  @param[in]     mailBoxHandle mailBox handle
 *  @param[in]     *buf          Mailbox Payload buffer pointer
 *  @param[in]     size          Mailbox Payload buffer size
 *  @param[in]     trans_id      transaction ID for the mail
 *  @retval        0 for success, -1 for failure -2 FULL
 *  @pre  
 *  @post 
 */
  XDAS_Int32 (*mailBoxWrite) (XDAS_Void *mailBoxHandle, XDAS_UInt8 *buf, 
               XDAS_UInt32 size, XDAS_UInt32 trans_id);

/**
 *  @brief Function mailBox_read() Reads from a mailBox. This is a non blocking 
 *                                 call
 *  @param[in]     mailBoxHandle mailBox handle
 *  @param[in]     *buf          Mailbox Payload buffer pointer
 *  @param[in]     *size         Mailbox Payload buffer size
 *  @param[in]     *trans_id     transaction ID for the mail
 *  @retval        0 for success, -1 for failure -3 Empty
 *  @pre  
 *  @post 
 */
  XDAS_Int32 (*mailBoxRead) (XDAS_Void *mailBoxHandle, XDAS_UInt8 *buf, 
               XDAS_UInt32 *size, XDAS_UInt32 *trans_id);

/**
 *  @brief Function mailBox_query() Polls mailBoxes for any available messages. 
                                    Non-blocking
 *  @param[in]     mailBoxHandle mailBox handle
 *  @retval        Number of messages in mailbox; negative error on failure
 *  @pre  
 *  @post 
 */
  XDAS_Int32 (*mailBoxQuery) (XDAS_Void *mailBoxHandle);

/** Identification parameters, to be set by the framework */
  XDAS_Int32 num_users;           /* Number of cores in the team - i.e. 
                             cores running together processing same stream */
  IVIDMC3_TASK_e task_ID;         /* Core task Identification: Reserved 
                                     MASTER_CORE */
  XDAS_Int32 user_id;             /* Core ID, include core inside chip and chip 
                                     ID */
} IVIDMC3_t;
#endif

/* Nothing past this point */
