/**
 * Contains main() function and SYS/BIOS start routines.
 */

#include <xdc/runtime/Log.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/IHeap.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/family/c66/Cache.h>
#include <ti/sysbios/heaps/HeapMem.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sdo/fc/rman/rman.h>
//#include <ti/csl/csl_bootcfgAux.h>

#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/HeapMemMP.h>
#include <ti/ipc/MessageQ.h>

#include "Heaps.h"
#include "Configuration.h"
#include "GlobalGates.h"
#include "Common.h"
#include "TranscodeComponent.h"
#include "EdmaResourceManagerConfig.h"

extern cregister volatile unsigned int DNUM;

#define DEVICE_REG32_W(x,y) *(volatile uint32_t *)(x)=(y)
#define IPCGR(x)            (0x02620240 + x*4)

extern void decodeMasterTask(UArg arg0, UArg arg1);
extern void decodeSlaveTask(UArg arg0, UArg arg1);

static ResultCode initializeRman()
{
    int edmaInstanceId = (DNUM & 1) + 1; // Balancing of EDMA instances (1 and 2) between cores
    int coreId = DNUM;
    int queueId = 0;

    System_printf("Using EDMA3 instance %d\n", edmaInstanceId);

    // Specify EDMA instance ID and queue ID
    EDMA3_PARAMS.regionConfig = &C6608_config[edmaInstanceId][coreId];
    ti_sdo_fc_edma3_EDMA3_physicalId = edmaInstanceId;
    ti_sdo_fc_edma3_Settings_defaultQ = queueId;

    int result = RMAN_init();

    if (result != IRES_OK) {
        System_printf("RMAN initialization failed (%d)\n", result);
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static void initializeCaching()
{
    Cache_Size size;
    size.l1pSize  = Cache_L1Size_32K; // L1P cache size
    size.l1dSize  = Cache_L1Size_32K; // L1D cache size
    size.l2Size   = Cache_L2Size_64K; // L2  cache size
    Cache_setSize(&size);

    UInt32 mar = Cache_getMar((Ptr) ddr3HeapRegionStartAddress) | ti_sysbios_family_c66_Cache_PC | ti_sysbios_family_c66_Cache_PFX;
    Cache_setMar((Ptr) ddr3HeapRegionStartAddress, ddr3HeapRegionSize, Cache_Mar_ENABLE | mar);

    const UInt32 disablePrefetch = (~ti_sysbios_family_c66_Cache_PFX) & (0xFFFFFFFF);
    const UInt32 disableCache = (~ti_sysbios_family_c66_Cache_PC) & (0xFFFFFFFF);
    mar = Cache_getMar((Ptr) ddr3UncachedHeapRegionStartAddress) & disableCache & disablePrefetch;
    Cache_setMar((Ptr) ddr3UncachedHeapRegionStartAddress, ddr3UncachedHeapRegionSize, Cache_Mar_DISABLE | mar);

    Cache_wbInvAll();
}

static ResultCode createSharedHeap(IHeap_Handle* heapHandle, int heapSize, char* heapName, int regionId, int heapId)
{
    if (DNUM == 0) {
        // Create heap

        HeapMemMP_Params params;
        HeapMemMP_Params_init(&params);

        params.regionId = regionId;
        params.name = heapName;
        params.sharedBufSize = heapSize - 0x180; // Minus heap creation overhead

        *heapHandle = (IHeap_Handle) HeapMemMP_create(&params);
        if (*heapHandle == NULL) {
            System_printf("Failed to create shared heap\n");
            return ResultCode_Error;
        }
    } else {
        // Open heap

        while (1) {
            if (HeapMemMP_open(heapName, (HeapMemMP_Handle*) heapHandle) == HeapMemMP_S_SUCCESS) {
                break;
            }
        }
    }

    if (MessageQ_registerHeap(*heapHandle, heapId) != MessageQ_S_SUCCESS) {
        System_printf("Failed to create shared heap\n");
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static ResultCode processConfiguration()
{
    int i;
    Error_Block eb;
    const DspConfiguration* configuration = Configuration_get();

    Error_init(&eb);

    // Initialize core statistics data

    int decodersCount = 0;

    for (i = 0; i < configuration->decodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->decoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            decodersCount++;
        }
    }

    for (i = 0; i < configuration->decodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->decoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            System_printf("Creating decoder task for decoder #%d\n", i);
            System_flush();

            Task_Params taskParams;
            Task_Params_init(&taskParams);
            taskParams.arg0 = i;
            taskParams.arg1 = 0;
            Task_FuncPtr taskFunction;

            if (coreIndex == 0) { // Master decoder
                taskFunction = decodeMasterTask;
                taskParams.stackSize = 0x2500;
            } else { // Slave decoder
                taskFunction = decodeSlaveTask;
                taskParams.stackSize = 0x2000;
            }

            Task_Handle taskHandle = Task_create(taskFunction, &taskParams, &eb);

            if (taskHandle == NULL) {
                System_printf("Failed to create task\n");
                return ResultCode_Error;
            }
        }
    }

    System_printf("Configuration processed\n");
    return ResultCode_Ok;
}

static ResultCode attachIpc()
{
    int i;

    for (i = 0; i < MultiProc_getNumProcessors(); i++) {
        if (i == MultiProc_self())  {
            continue;
        }

        System_printf("IPC: attaching to core#%d\n", i);
        System_flush();

        int result;
        do {
            result = Ipc_attach(i);
        } while (result == Ipc_E_NOTREADY);

        if (result != Ipc_S_SUCCESS) {
            return ResultCode_Error;
        }

        System_printf("Attached\n");
    }

    return ResultCode_Ok;
}

static ResultCode startIpc()
{
    // Set localId
    char procName[5] = "COREx";
    procName[4] = '0' + DNUM;
    MultiProc_setLocalId(MultiProc_getId(procName));

    // Start IPC
    int ret = Ipc_start();
    if (ret != Ipc_S_SUCCESS) {
        System_printf("Failed to initialize IPC (%d)\n", ret);
        return ResultCode_Error;
    }

    if (attachIpc() != ResultCode_Ok) {
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static void initializeTask(UArg arg0, UArg arg1)
{
    System_printf("Initializing GateMP\n");

    if (GlobalGates_init() != ResultCode_Ok) {
        System_printf("Failed to initialize global gates\n");
        goto INITIALIZE_FAILED;
    }

    System_printf("Creating MSMC heap\n");

    if (createSharedHeap(&msmcHeap, msmcHeapRegionSize, "msmcHeap", 1, 1) != ResultCode_Ok) {
        goto INITIALIZE_FAILED;
    }

    System_printf("Creating DDR3 heap\n");

    if (createSharedHeap(&ddr3Heap, ddr3HeapRegionSize, "ddr3Heap", 2, 2) != ResultCode_Ok) {
        goto INITIALIZE_FAILED;
    }

    System_printf("Creating DDR3 uncached heap\n");

    if (createSharedHeap(&ddr3UncachedHeap, ddr3UncachedHeapRegionSize, "ddr3UncachedHeap", 3, 3) != ResultCode_Ok) {
        goto INITIALIZE_FAILED;
    }

    Heaps_printState();

#if RUN_ON_EVM == 0
    if (Control_create((void*) Configuration_get()->controlBufferAddresses[DNUM], Configuration_get()->reportingBufferSizes[DNUM]) != ResultCode_Ok) {
        System_printf("Failed to create message queue\n");
        goto INITIALIZE_FAILED;
    }
#endif

    System_printf("Initializing RMAN\n");

    if (initializeRman() != ResultCode_Ok) {
        goto INITIALIZE_FAILED;
    }

    System_printf("Initializing resize component\n");

    if (DNUM == 0) {
        TranscodeComponent_initModule();
    }

    System_printf("Processing configuration\n");

    if (processConfiguration() != ResultCode_Ok) {
        goto INITIALIZE_FAILED;
    }

    System_flush();
    return;

INITIALIZE_FAILED:

    System_flush();
    BIOS_exit(0);
}

void main()
{
    const DspConfiguration* configuration = Configuration_get();

    if (configuration->marker != CONFIGURATION_MARKER) { // Wrong configuration
        goto END;
    }

#if USE_DEFAULT_CIO == 0
    if (Reporting_initializeCio((void*) configuration->reportingBufferAddresses[DNUM], configuration->reportingBufferSizes[DNUM]) != ResultCode_Ok) {
        goto END;
    }
#endif

    System_printf("main() on CORE=%d\n", DNUM);
    System_flush();

#if KICK_SLAVE_CORES != 0
    if (DNUM == 0) { // Kick other cores
        // Unlock the chip registers
        CSL_BootCfgUnlockKicker();

        // Wake up the other core to run memory test
        int core;

        for (core = 1; core < CORES_PER_DSP; core++) {
            System_printf("Activating core %d\n", core);
            // IPC interrupt other cores
            DEVICE_REG32_W(IPCGR(core), 1);
        }
    }
#endif

    System_printf("Configuring caching\n");

    initializeCaching();

    System_printf("Starting IPC\n");

    if (startIpc() != ResultCode_Ok) {
        goto END;
    }

    System_printf("Starting SYS/BIOS\n");
    System_flush();

    Error_Block eb;
    Error_init(&eb);
    Task_Params taskParams;
    Task_Params_init(&taskParams);
    taskParams.stackSize = 0x1000;
    taskParams.arg0 = 0;
    taskParams.arg1 = 0;

    if (Task_create(initializeTask, &taskParams, &eb) == NULL) {
        System_printf("Failed to create initialization task\n");
        goto END;
    }

    BIOS_start();

END:

    System_flush();
    BIOS_exit(0);
}
